package lcsb.mapviewer.converter.model.sbml.species;

import java.awt.Color;
import java.util.Objects;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.sbml.jsbml.Model;
import org.sbml.jsbml.ext.layout.BoundingBox;
import org.sbml.jsbml.ext.layout.Dimensions;
import org.sbml.jsbml.ext.layout.Point;
import org.sbml.jsbml.ext.render.ColorDefinition;
import org.sbml.jsbml.ext.render.Ellipse;
import org.sbml.jsbml.ext.render.HTextAnchor;
import org.sbml.jsbml.ext.render.LocalStyle;
import org.sbml.jsbml.ext.render.Polygon;
import org.sbml.jsbml.ext.render.Rectangle;
import org.sbml.jsbml.ext.render.RelAbsVector;
import org.sbml.jsbml.ext.render.RenderPoint;
import org.sbml.jsbml.ext.render.VTextAnchor;

import lcsb.mapviewer.converter.model.sbml.SbmlModelUtils;
import lcsb.mapviewer.model.LogMarker;
import lcsb.mapviewer.model.ProjectLogEntryType;
import lcsb.mapviewer.model.graphics.ArrowType;
import lcsb.mapviewer.model.map.species.field.BindingRegion;
import lcsb.mapviewer.model.map.species.field.CodingRegion;
import lcsb.mapviewer.model.map.species.field.ModificationResidue;
import lcsb.mapviewer.model.map.species.field.ModificationSite;
import lcsb.mapviewer.model.map.species.field.ProteinBindingDomain;
import lcsb.mapviewer.model.map.species.field.RegulatoryRegion;
import lcsb.mapviewer.model.map.species.field.Residue;
import lcsb.mapviewer.model.map.species.field.StructuralState;
import lcsb.mapviewer.model.map.species.field.TranscriptionSite;
import lcsb.mapviewer.modelutils.map.ElementUtils;

public class ModificationResidueExporter {

  private static Logger logger = LogManager.getLogger();

  private SbmlModelUtils sbmlModelUtils;

  public ModificationResidueExporter(final Model sbmlModel) {
    this.sbmlModelUtils = new SbmlModelUtils(sbmlModel);
  }

  public LocalStyle createStyle(final ModificationResidue element) {
    if (element instanceof Residue) {
      return createEllipseStyle(element);
    } else if (element instanceof ModificationSite) {
      return createModificationSiteStyle((ModificationSite) element);
    } else if (element instanceof TranscriptionSite) {
      return createTranscriptionSiteStyle((TranscriptionSite) element);
    } else if (element instanceof RegulatoryRegion) {
      return createRectangleStyle(element);
    } else if (element instanceof BindingRegion) {
      return createRectangleStyle(element);
    } else if (element instanceof CodingRegion) {
      return createRectangleStyle(element);
    } else if (element instanceof ProteinBindingDomain) {
      return createRectangleStyle(element);
    } else if (element instanceof StructuralState) {
      return createEllipseStyle(element);
    } else {
      logger.warn(new LogMarker(ProjectLogEntryType.EXPORT_ISSUE, element.getSpecies()),
          "Don't know how to render modification residue: " + new ElementUtils().getElementTag(element));
      return createEllipseStyle(element);
    }
  }

  private LocalStyle createRectangleStyle(final ModificationResidue element) {
    LocalStyle style = createGenericStyle(element);

    Rectangle rectangle = new Rectangle();
    rectangle.setRx(new RelAbsVector(0, 0));
    rectangle.setRy(new RelAbsVector(0, 0));
    rectangle.setX(new RelAbsVector(0, 0));
    rectangle.setY(new RelAbsVector(0, 0));
    rectangle.setWidth(new RelAbsVector(0, 100));
    rectangle.setHeight(new RelAbsVector(0, 100));

    style.getGroup().addElement(rectangle);

    return style;
  }

  private LocalStyle createTranscriptionSiteStyle(final TranscriptionSite transcriptionSite) {
    LocalStyle style = createGenericStyle(transcriptionSite);

    ColorDefinition color = sbmlModelUtils.getColorDefinition(transcriptionSite.getBorderColor());

    style.getGroup().setStroke(color.getId());
    style.getGroup().setEndHead(ArrowType.FULL.name());
    style.getGroup().setFill(color.getId());

    double y = transcriptionSite.getY();
    double x = transcriptionSite.getX();
    if (Objects.equals(transcriptionSite.getDirection(), "RIGHT")) {
      x += transcriptionSite.getWidth();
    }
    Polygon line = new Polygon();

    RenderPoint p1 = line.createRenderPoint();
    p1.setX(new RelAbsVector(Math.round(x - transcriptionSite.getX())));
    p1.setY(new RelAbsVector(Math.round(y + transcriptionSite.getHeight() - transcriptionSite.getY())));

    RenderPoint p2 = line.createRenderPoint();
    p2.setX(new RelAbsVector(Math.round(x - transcriptionSite.getX())));
    p2.setY(new RelAbsVector(Math.round(y - transcriptionSite.getY())));

    double delta = 0;
    if (Objects.equals(transcriptionSite.getDirection(), "RIGHT")) {
      x -= transcriptionSite.getWidth();
      delta = 6;
    } else {
      x += transcriptionSite.getWidth();
      delta = -6;
    }
    RenderPoint p3 = line.createRenderPoint();
    p3.setX(new RelAbsVector(Math.round(x - transcriptionSite.getX())));
    p3.setY(new RelAbsVector(Math.round(y - transcriptionSite.getY())));

    RenderPoint p4 = line.createRenderPoint();
    p4.setX(new RelAbsVector(Math.round(x + delta - transcriptionSite.getX())));
    p4.setY(new RelAbsVector(Math.round(y - 3 - transcriptionSite.getY())));

    RenderPoint p5 = line.createRenderPoint();
    p5.setX(new RelAbsVector(Math.round(x + delta - transcriptionSite.getX())));
    p5.setY(new RelAbsVector(Math.round(y + 3 - transcriptionSite.getY())));

    RenderPoint p6 = line.createRenderPoint();
    p6.setX(new RelAbsVector(Math.round(x - transcriptionSite.getX())));
    p6.setY(new RelAbsVector(Math.round(y - transcriptionSite.getY())));

    RenderPoint p7 = line.createRenderPoint();
    p7.setX(new RelAbsVector(p2.getX()));
    p7.setY(new RelAbsVector(p2.getY()));

    style.getGroup().addElement(line);

    return style;
  }

  private LocalStyle createModificationSiteStyle(final ModificationSite modificationSite) {
    LocalStyle style = createGenericStyle(modificationSite);

    Ellipse ellipse = new Ellipse();
    ellipse.setCx(new RelAbsVector(0, 50));
    ellipse.setCy(new RelAbsVector(Math.round(modificationSite.getWidth() / 2), 0));
    ellipse.setRx(new RelAbsVector(0, 50));
    ellipse.setRy(new RelAbsVector(Math.round(modificationSite.getWidth() / 2), 0));

    style.getGroup().addElement(ellipse);

    Polygon line = new Polygon();

    RenderPoint p1 = line.createRenderPoint();
    p1.setX(new RelAbsVector(0, 50));
    p1.setY(new RelAbsVector(modificationSite.getWidth()));

    RenderPoint p2 = line.createRenderPoint();
    p2.setX(new RelAbsVector(0, 50));
    p2.setY(new RelAbsVector(0, 100));

    style.getGroup().addElement(line);

    return style;
  }

  private LocalStyle createEllipseStyle(final ModificationResidue element) {
    LocalStyle style = createGenericStyle(element);

    Ellipse ellipse = new Ellipse();
    ellipse.setCx(new RelAbsVector(0, 50));
    ellipse.setCy(new RelAbsVector(0, 50));
    ellipse.setRx(new RelAbsVector(0, 50));
    ellipse.setRy(new RelAbsVector(0, 50));

    style.getGroup().addElement(ellipse);

    return style;
  }

  private LocalStyle createGenericStyle(final ModificationResidue element) {
    LocalStyle style = createGenericStyle();
    style.getGroup().setStroke(sbmlModelUtils.getColorDefinition(element.getBorderColor()).getId());
    style.getGroup().setFontSize(element.getFontSize().shortValue());
    return style;
  }

  private LocalStyle createGenericStyle() {
    LocalStyle style = sbmlModelUtils.createStyle();
    ColorDefinition color = sbmlModelUtils.getColorDefinition(Color.BLACK);
    ColorDefinition background = sbmlModelUtils.getColorDefinition(Color.WHITE);
    style.getGroup().setStroke(color.getId());
    style.getGroup().setStrokeWidth(1.0);
    style.getGroup().setFill(background.getId());
    style.getGroup().setFontSize((short) 10);
    return style;
  }

  public LocalStyle createTextStyle(final ModificationResidue mr) {
    return createTextStyle();
  }

  private LocalStyle createTextStyle() {
    LocalStyle style = createGenericStyle();

    style.getGroup().setVTextAnchor(VTextAnchor.MIDDLE);
    style.getGroup().setTextAnchor(HTextAnchor.MIDDLE);
    return style;
  }

  public LocalStyle createTextStyle(final StructuralState structuralState) {
    return createTextStyle();
  }

  public BoundingBox createBoundingBoxForModification(final ModificationResidue mr) {
    double width = mr.getWidth();
    double height = mr.getHeight();
    double x = mr.getX();
    double y = mr.getY();
    double z = mr.getZ();
    BoundingBox boundingBox = new BoundingBox();
    boundingBox.setPosition(new Point(x, y, z));
    Dimensions dimensions = new Dimensions();
    dimensions.setWidth(width);
    dimensions.setHeight(height);
    boundingBox.setDimensions(dimensions);
    return boundingBox;
  }

  public BoundingBox createBoundingBoxForModificationText(final ModificationResidue mr) {
    BoundingBox result = createBoundingBoxForModification(mr);
    result.getPosition().setX(result.getPosition().getX() + 2);
    result.getPosition().setY(result.getPosition().getY() + 2);
    result.getDimensions().setWidth(result.getDimensions().getWidth() - 4);
    if (mr instanceof ModificationSite) {
      result.getDimensions().setHeight(result.getDimensions().getWidth());
    } else {
      result.getDimensions().setHeight(result.getDimensions().getHeight() - 4);
    }
    result.getPosition().setZ(result.getPosition().getZ() + 1);
    return result;
  }

  public BoundingBox createBoundingBoxForStructuralState(final StructuralState structuralState) {
    double width = structuralState.getWidth();
    double height = structuralState.getHeight();
    double x = structuralState.getX();
    double y = structuralState.getY();
    BoundingBox boundingBox = new BoundingBox();
    boundingBox.setPosition(new Point(x, y, structuralState.getZ()));
    Dimensions dimensions = new Dimensions();
    dimensions.setWidth(width);
    dimensions.setHeight(height);
    boundingBox.setDimensions(dimensions);
    return boundingBox;
  }

  public BoundingBox createBoundingBoxForStructuralStateText(final StructuralState structuralState) {
    BoundingBox result = createBoundingBoxForStructuralState(structuralState);
    result.getPosition().setZ(result.getPosition().getZ() + 1);
    return result;
  }

}
