"use strict";
require("../../../mocha-config");

// noinspection JSUnusedLocalSymbols
var logger = require('../../../logger');

var GoogleMapsApiCanvas = require('../../../../../main/js/map/canvas/GoogleMaps/GoogleMapsApiCanvas');
var Point = require('../../../../../main/js/map/canvas/Point');
var Bounds = require('../../../../../main/js/map/canvas/Bounds');

var chai = require('chai');
var assert = chai.assert;

describe('GoogleMapsApiCanvas', function () {
  var testOptions = {
    center: new Point(0, 0),
    tileSize: 256,
    width: 300,
    height: 600,
    minZoom: 2,
    zoom: 5,
    backgroundOverlays: [{
      id: 1,
      name: "overlay",
      directory: "overlay_dir"
    }]
  };
  it("constructor", function () {
    var canvas = new GoogleMapsApiCanvas(testDiv, testOptions);
    assert.ok(canvas.pixelOrigin_);
    assert.ok(canvas.pixelsPerLonDegree_);
    assert.ok(canvas.pixelsPerLonRadian_);
    assert.ok(canvas.zoomFactor);

  });

  it("latLngToTile", function () {
    var canvas = new GoogleMapsApiCanvas(testDiv, testOptions);

    var latLng = new google.maps.LatLng(0, 1);

    var tileCoordinates = canvas.latLngToTile(latLng, 3);
    assert.ok(tileCoordinates);
    assert.ok(tileCoordinates instanceof Point);

  });

  describe("createMapTypeOption", function () {
    it("default", function () {
      var canvas = new GoogleMapsApiCanvas(testDiv, testOptions);

      var overlayDataObj = testOptions.backgroundOverlays[0];

      var result = canvas.createMapTypeOption(overlayDataObj);

      assert.ok(result);

    });
    it("no background overlays", function () {
      var options = {
        center: new Point(0, 0),
        tileSize: 256,
        width: 300,
        height: 600,
        minZoom: 2,
        zoom: 5,
        backgroundOverlays: []
      };

      var canvas = new GoogleMapsApiCanvas(testDiv, options);
      assert.ok(canvas);
    });
    it("getTileUrl", function () {
      var coordinates = {x: 0, y: 0};
      var canvas = new GoogleMapsApiCanvas(testDiv, testOptions);

      var overlayDataObj = testOptions.backgroundOverlays[0];

      var typeOptions = canvas.createMapTypeOption(overlayDataObj);
      var url = typeOptions.getTileUrl(coordinates, 3);
      assert.ok(url);
      assert.ok(url.indexOf(coordinates.x) >= 0);
      assert.ok(url.indexOf(coordinates.y) >= 0);
    });
  });


  it("Point - Lat,Lng conversion", function () {

    var canvas = new GoogleMapsApiCanvas(testDiv, testOptions);

    var x = 199;
    var y = 110;
    var point = new Point(x, y);

    var coordinates = canvas.fromPointToLatLng(point);

    var point2 = canvas.fromLatLngToPoint(coordinates);

    assert.closeTo(point2.x, point.x, 0.5, "X coordinate is invalid after transformation");
    assert.closeTo(point2.y, point.y, 0.5, "Y coordinate is invalid after transformation");

  });

  it("click on rectangle", function () {
    var canvas = new GoogleMapsApiCanvas(testDiv, testOptions);
    var rectangle = canvas.createRectangle({
      fillOpacity: 0.7,
      strokeColor: "#330000",
      strokeOpacity: 0.5,
      strokeWeight: 1.0,
      id: "1",
      fillColor: "#00FF00",
      bounds: new Bounds(new Point(0, 0), new Point(2, 2))
    });
    return google.maps.event.trigger(rectangle.getGoogleRectangles()[0], "click");
  });

  it("set rectangle bounds", function () {
    var canvas = new GoogleMapsApiCanvas(testDiv, testOptions);
    var rectangle = canvas.createRectangle({
      fillOpacity: 0.7,
      strokeColor: "#330000",
      strokeOpacity: 0.5,
      strokeWeight: 1.0,
      id: "1",
      fillColor: "#00FF00",
      bounds: new Bounds(new Point(10, 10), new Point(20, 200))
    });
    rectangle.setBounds(new Bounds(new Point(100, 10), new Point(150, 200)));
    var newBounds = rectangle.getBounds();
    assert.closeTo(newBounds.getTopLeft().distanceTo(new Point(100, 10)), 0, helper.EPSILON);
    assert.closeTo(newBounds.getRightBottom().distanceTo(new Point(150, 200)), 0, helper.EPSILON);
  });

  it("create rectangle with gradient", function () {
    var canvas = new GoogleMapsApiCanvas(testDiv, testOptions);
    var rectangle = canvas.createRectangle({
      fillOpacity: 0.7,
      strokeColor: "#330000",
      strokeOpacity: 0.5,
      strokeWeight: 1.0,
      id: "1",
      fillGradient: [{color: "#00FF00", amount: 1}, {color: "#00aa00", amount: 5}],
      bounds: new Bounds(new Point(10, 10), new Point(20, 200))
    });
  });

});
