package lcsb.mapviewer.services.interfaces;

import lcsb.mapviewer.commands.CommandExecutionException;
import lcsb.mapviewer.converter.ConverterException;
import lcsb.mapviewer.converter.InvalidInputDataExecption;
import lcsb.mapviewer.converter.graphics.DrawingException;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.ProjectLogEntry;
import lcsb.mapviewer.model.ProjectLogEntryType;
import lcsb.mapviewer.model.ProjectStatus;
import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.model.job.MinervaJob;
import lcsb.mapviewer.model.job.MinervaJobExecutor;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.layout.ProjectBackground;
import lcsb.mapviewer.model.overlay.InvalidDataOverlayException;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.model.user.UserAnnotationSchema;
import lcsb.mapviewer.services.ObjectNotFoundException;
import lcsb.mapviewer.services.utils.CreateProjectParams;
import org.springframework.transaction.annotation.Transactional;

import java.io.IOException;
import java.util.List;
import java.util.Set;

/**
 * Service that manages projects.
 *
 * @author Piotr Gawron
 */
public interface IProjectService extends MinervaJobExecutor {

  /**
   * Returns a project with a give {@link Project#getProjectId()  project identifier}.
   *
   * @param projectId {@link Project#getProjectId()}  project identifier}
   * @return project with an identifier given as parameter. Null if such project  doesn't exist.
   */
  @Transactional
  Project getProjectByProjectId(final String projectId);

  @Transactional
  Project getProjectByProjectId(final String projectId, final boolean initializeLazy);

  /**
   * Checks if project with a given {@link Project#getProjectId()  identifier} exists.
   *
   * @param projectId {@link Project#getProjectId()}  project identifier}
   * @return <code>true</code> if the project with the given name exists, <code>false</code> otherwise
   */
  @Transactional
  boolean projectExists(final String projectId);

  /**
   * Returns list of all projects.
   *
   * @return list of all projects.
   */
  @Transactional
  List<Project> getAllProjects();

  @Transactional
  List<Project> getAllProjects(boolean initializeLazy);

  /**
   * Removes project from the system.
   *
   * @param project object to remove
   */
  @Transactional
  MinervaJob submitRemoveProjectJob(final Project project);

  /**
   * Creates project using give parameters. See {@link CreateProjectParams}.
   *
   * @param params information about project to create
   */
  @Transactional
  MinervaJob submitCreateProjectJob(final CreateProjectParams params) throws SecurityException;

  /**
   * Method that updates information about raw {@link Project} object.
   *
   * @param project project to update
   */
  @Transactional
  void updateProject(final Project project);

  @Transactional
  UserAnnotationSchema prepareUserAnnotationSchema(final User user, final boolean initializeLazy);

  @Transactional
  void removeBackground(final ProjectBackground projectBackground);

  @Transactional
  void removeBackground(int backgroundId, String path);

  @Transactional
  ProjectBackground getBackgroundById(final String projectId, final Integer backgroundId, final boolean initializeLazy);

  @Transactional
  void updateBackground(final ProjectBackground background);

  @Transactional
  long getNextId();

  @Transactional
  UploadedFileEntry getFileByProjectId(final String projectId);

  @Transactional
  List<ProjectBackground> getBackgrounds(final String projectId, final boolean initializeLazy);

  @Transactional
  void delete(final Project project);

  @Transactional
  void add(final Project project);

  @Transactional
  void addLogEntry(final Project p, final ProjectLogEntry entry);

  @Transactional
  MinervaJob projectFailure(final String projectId, final ProjectLogEntryType errorType, final String message, final Throwable e);

  @Transactional
  void removeProject(String projectId, String directory);

  @Transactional
  Project createModel(CreateProjectParams params)
      throws InvalidInputDataExecption, ConverterException, IOException, InvalidDataOverlayException, InstantiationException, IllegalAccessException,
      CommandExecutionException;

  @Transactional
  void analyzeAnnotations(String projectId, User user);

  /**
   * This method creates set of images for the model backgrounds.
   *
   * @param params configuration parameters including set of backgrounds to generate
   * @throws IOException               thrown when there are problems with generating files
   * @throws DrawingException          thrown when there was a problem with drawing a map
   * @throws CommandExecutionException thrown when one of the files describing backgrounds is invalid
   */
  @Transactional
  void createImages(CreateProjectParams params) throws IOException, DrawingException, CommandExecutionException;

  @Transactional
  void cacheData(String projectId) throws ObjectNotFoundException;

  /**
   * Sends email about unsuccessful project creation.
   *
   * @param projectName name of the project
   * @param email       email where we want to send information
   * @param e           exception that caused problem
   */
  @Transactional
  void sendUnsuccesfullEmail(String projectName, String email, Throwable e);

  /**
   * Updates status of the generating project.
   *
   * @param projectId project that is generated
   * @param status    what is the current status
   * @param progress  what is the progress
   */
  @Transactional
  void updateProjectStatus(String projectId, ProjectStatus status, double progress);

  @Transactional
  void addLogEntries(String projectId, Set<ProjectLogEntry> createLogEntries);

  @Transactional
  void submitArchiveProjectJob(String projectId);

  String getProjectHomeDir(Project project);

  @Transactional
  void submitReviveProjectJob(String projectId);

  @Transactional
  void reviveBackgrounds(String projectId);

  @Transactional
  Set<MiriamData> getMiriamForProject(String projectId);

  @Transactional
  List<MiriamData> getMiriamWithDuplicatesForProject(String projectId);

}
