package lcsb.mapviewer.services.search.chemical;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;

import javax.annotation.PostConstruct;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.annotation.data.Chemical;
import lcsb.mapviewer.annotation.services.annotators.AnnotatorException;
import lcsb.mapviewer.annotation.services.annotators.HgncAnnotator;
import lcsb.mapviewer.annotation.services.dapi.ChemicalParser;
import lcsb.mapviewer.annotation.services.dapi.ChemicalSearchException;
import lcsb.mapviewer.common.IProgressUpdater;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.common.exception.NotImplementedException;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.job.MinervaJob;
import lcsb.mapviewer.model.job.MinervaJobType;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.Rna;
import lcsb.mapviewer.services.interfaces.IMinervaJobService;
import lcsb.mapviewer.services.jobs.RefreshChemicalInfoMinervaJob;
import lcsb.mapviewer.services.search.DbSearchCriteria;
import lcsb.mapviewer.services.search.DbSearchService;

/**
 * Implementation of the service that allows access information about chemicals.
 * 
 * @author Ayan Rota
 * 
 */
@Transactional
@Service
public class ChemicalService extends DbSearchService implements IChemicalService {

  /**
   * Default class logger.
   */
  private Logger logger = LogManager.getLogger();

  /**
   * Access point and parser for the online ctd database.
   */
  private ChemicalParser chemicalParser;

  /**
   * Object used to perform operations on {@link MiriamType#HGNC} annotations.
   */
  private HgncAnnotator hgncAnnotator;

  private IMinervaJobService minervaJobService;

  @Autowired
  private IChemicalService self;

  @Autowired
  public ChemicalService(final ChemicalParser chemicalParser,
      final HgncAnnotator hgncAnnotator,
      final IMinervaJobService minervaJobService) {
    super();
    this.chemicalParser = chemicalParser;
    this.hgncAnnotator = hgncAnnotator;
    this.minervaJobService = minervaJobService;
  }

  @Override
  public Chemical getByName(final String name, final DbSearchCriteria searchCriteria) {
    if (searchCriteria.getDisease() == null) {
      throw new InvalidArgumentException("diseaseID cannot be null");
    }
    if (name == null || name.isEmpty()) {
      return null;
    }
    try {
      List<Chemical> result = chemicalParser.getChemicalsByName(searchCriteria.getDisease(), name);
      if (result == null) {
        return null;
      }
      if (result.size() == 0) {
        return null;
      } else {
        if (result.size() > 1) {
          logger.warn("More than one chemical found by name: " + name);
        }
        Chemical chemical = result.get(0);
        addProjectTargets(searchCriteria.getProject(), chemical);
        return chemical;
      }
    } catch (final ChemicalSearchException e) {
      logger.error("Problem with fetch chemical data", e);
      return null;
    }
  }

  @Override
  public List<Chemical> getForTargets(final Collection<Element> targets, final DbSearchCriteria searchCriteria) {
    List<Chemical> chemicalList = new ArrayList<>();
    Set<MiriamData> targetsMiriam = new HashSet<>();
    for (final Element element : targets) {
      if (element instanceof Protein || element instanceof Gene || element instanceof Rna) {
        boolean hgncFound = false;
        for (final MiriamData md : element.getMiriamData()) {
          if (MiriamType.HGNC_SYMBOL.equals(md.getDataType())) {
            targetsMiriam.add(md);
            hgncFound = true;
          }
        }
        if (!hgncFound) {
          MiriamData md = new MiriamData(MiriamType.HGNC_SYMBOL, element.getName());
          try {
            if (hgncAnnotator.isValidHgncMiriam(md)) {
              targetsMiriam.add(md);
            }
          } catch (final AnnotatorException e) {
            logger.error("Problem with accessing HGNC database", e);
          }
        }
      }
    }
    try {
      chemicalList = chemicalParser.getChemicalListByTargets(targetsMiriam, searchCriteria.getDisease());
    } catch (final ChemicalSearchException e) {
      logger.error("Problem with accessing chemical database", e);
    }

    Collections.sort(chemicalList, new Chemical.NameComparator());
    for (final Chemical chemical : chemicalList) {
      addProjectTargets(searchCriteria.getProject(), chemical);
    }
    return chemicalList;
  }

  @PostConstruct
  public void init() {
    minervaJobService.registerExecutor(MinervaJobType.REFRESH_CHEMICAL_INFO, self);
  }

  @Override
  public void cacheDataForModel(final ModelData originalModel, final IProgressUpdater progressUpdater) {
    double progress = 0.0;
    progressUpdater.setProgress(progress * IProgressUpdater.MAX_PROGRESS);
    MiriamData disease = originalModel.getProject().getDisease();
    if (disease != null) {
      logger.debug("Caching chemical queries...");
      Set<MiriamData> targetMiriams = getHgncIdentifiers(originalModel);

      double counter = 0.0;
      for (final MiriamData md : targetMiriams) {
        refreshChemical(disease, md);
        counter += 1;
        progress = counter / targetMiriams.size();
        progressUpdater.setProgress(progress * IProgressUpdater.MAX_PROGRESS);
      }
      try {
        getSuggestedQueryList(originalModel.getProject(), originalModel.getProject().getOrganism());
      } catch (final ChemicalSearchException e) {
        logger.error("Problem with accessing chemical info", e);
      }

    } else {
      logger.debug("Caching chemical queries not possible...");
    }
  }

  private void refreshChemical(final MiriamData disease, final MiriamData md) {
    try {
      List<Chemical> chemicalList = chemicalParser.getChemicalListByTarget(md, disease);
      for (final Chemical chemical : chemicalList) {
        cacheMiriamData(chemical);
      }
    } catch (final ChemicalSearchException | AnnotatorException e) {
      logger.error("Problem with accessing info about chemical for target: " + md, e);
    }
  }

  @Override
  public List<String> getSuggestedQueryList(final Project project, final MiriamData disease) throws ChemicalSearchException {
    return chemicalParser.getSuggestedQueryList(project, disease);
  }

  @Override
  public void execute(final MinervaJob job) throws Exception {
    if (job.getJobType() == MinervaJobType.REFRESH_CHEMICAL_INFO) {
      RefreshChemicalInfoMinervaJob parametres = new RefreshChemicalInfoMinervaJob(job.getJobParameters());
      MiriamData md = new MiriamData(parametres.getType(), parametres.getId());
      MiriamData disease = new MiriamData(parametres.getDiseaseType(), parametres.getDiseaseId());
      if (Objects.equals(MiriamType.HGNC_SYMBOL, md.getDataType())) {
        refreshChemical(disease, md);
      }
    } else {
      throw new NotImplementedException(job.getJobType() + " not implemented");
    }
  }

}
